/*
 * Cree a partir de catMouse.cpp (voir le site de RL-Glue)
 * Auteur de cette version : Pierre Gérard pierre.gerard@lipn.univ-paris13.fr
 *
 * Environnement de type Gridworld avec des obstacles, des trous de souris, un chat et du fromage
 * Le chat ne peut pas accéder aux cases "trou de souris"
 * Le chat est egreedy en regard du rapprochement de la souris
 *
 * Observations : int*
 * - int[0] : colonne de l'agent souris
 * - int[1] : ligne de l'agent souris
 * - int[2] : colonne du chat
 * - int[3] : ligne du chat
 * Actions : int
 * - 0..7 pour les huit directions
 * Récompense : float, typiquement, on met
 * - -10 si la souris et le chat se trouvent sur la meme case
 * - +5 si la souris se trouve sur une case fromage
 *
 * Tache episodique : un épisode se termine quand le chat et la souris sont sur la meme case (RIP la souris) 
 * Chaque episode, le point de depart est choisi aleatoirement
 */

#include <ctime>
#include "Catmouse.h"

long step_count;
long episode_count;

int random_cheese = false;    	// est-ce que le placement du fromage est aléatoire ?
				// à n'utiliser qu'avec les environnements sans fromage

long verbose_env_from = 1490; 	// épisode à partir duquel l'environement devient verbose

int tempo = 50000;		// temporisation enre deux pas de temps, dans le cas verbose

Task_specification env_init()
{
	cheeseReward=5;
	deathPenalty=-10;
	epsilon_cat = 0.5;	// pourcentage de chances de provoquer une action valide aléatoire

	//0 : vide
	//1 : obstacle
	//2 : trou de souris
	//3 : fromage

	// ENVIRONNEMENTS D'ESSAI (essayez les vôtres !)

// 	row = 3 ; col = 7;
// 	int laby[3][7]={	{0,0,0,3,0,0,0},
// 				{0,1,1,1,1,1,0},
// 				{0,0,0,3,0,0,0}};

	row = 3 ; col = 7;
	int laby[3][7]={	{0,0,0,0,0,0,0},
				{3,1,1,1,1,1,3},
				{0,0,0,0,0,0,0}};

// 	random_cheese=true;
// 	row = 3 ; col = 7;
// 	int laby[3][7]={	{0,0,0,0,0,0,0},
// 				{0,1,1,1,1,1,0},
// 				{0,0,0,0,0,0,0}};

// 	row = col = 6;
// 	int laby[6][6]={	{0,0,0,0,0,0},
// 				{0,3,1,1,3,0},
// 				{0,1,1,1,1,0},
// 				{0,1,1,1,1,0},
// 				{0,3,1,1,3,0},
// 				{0,0,0,0,0,0}};

// 	row = 5 ; col = 8;
// 	int laby[5][8]={	{0,0,0,0,0,0,0,0},
// 				{0,1,1,1,1,1,1,0},
// 				{3,0,0,0,0,0,0,3},
// 				{0,1,1,1,1,1,1,0},
// 				{0,0,0,0,0,0,0,0}};

// 	row = 9 ; col = 8;
// 	int laby[9][8]={	{0,0,0,0,0,0,0,0},
// 				{0,0,1,1,1,1,0,0},
// 				{0,1,1,1,1,1,1,0},
// 				{0,0,1,1,1,1,0,0},
// 				{3,0,0,0,0,0,0,3},
// 				{0,0,1,1,1,1,0,0},
// 				{0,1,1,1,1,1,1,0},
// 				{0,0,1,1,1,1,0,0},
// 				{0,0,0,0,0,0,0,0}};

// 	row = 8 ; col = 9;
// 	int laby[8][9]={	{0,0,1,0,3,0,1,0,0},
// 				{0,1,0,1,0,1,0,1,0},
// 				{0,1,0,1,0,1,0,1,0},
// 				{0,1,0,1,0,1,0,1,0},
// 				{0,1,0,1,0,1,0,1,0},
// 				{0,1,0,1,0,1,0,1,0},
// 				{0,1,0,1,0,1,0,1,0},
// 				{0,0,1,0,3,0,1,0,0}};

// 	row = 4 ; col = 9;
// 	int laby[4][9]={	{0,0,0,0,3,0,0,0,0},
// 				{0,0,0,0,0,0,0,0,0},
// 				{0,1,1,1,2,1,1,1,0},
// 				{0,0,0,0,0,0,0,0,0}};

// 	random_cheese = true;
// 	row = 9 ; col = 9;
// 	int laby[9][9]={	{0,0,0,0,0,0,0,0,0},
// 				{0,1,1,1,0,1,1,1,0},
// 				{0,1,0,0,0,0,0,1,0},
// 				{0,1,0,1,1,1,0,1,0},
// 				{0,0,0,1,1,1,0,0,0},
// 				{0,1,0,1,1,1,0,1,0},
// 				{0,1,0,0,0,0,0,1,0},
// 				{0,1,1,1,0,1,1,1,0},
// 				{0,0,0,0,0,0,0,0,0}};


	walls = new int*[row];
	for( int i = 0; i < row; i++)
		walls[i] = new int[col];
	
	for(int i=0;i< row; i++)
		for(int j=0;j< col; j++)
			walls[i][j] = laby[i][j];

	if (random_cheese) setRandomCheese();

	static char ts[1024];

	sprintf(ts,"1.1:e:4_[i,i,i,i]_[0,%d]_[0,%d]_[0,%d]_[0,%d]:1_[i]_[0,7]", row-1,col-1,row-1,col-1);

	env_print();

	step_count=0;
	episode_count=0;

	return ts;
}
	

Observation env_start()
{
		step_count++; 
		episode_count++;

		setRandomPos();

		o[0] = mx;
		o[1] = my;
		o[2] = cx;
		o[3] = cy;

		if (episode_count>verbose_env_from) env_print(); else printf("*");

		return o;
}

Reward_observation env_step(Action action) 
{
		step_count++;

		//move mouse based on action
		Dimension d;

		d = moveMouse(action);
		int x=d.width, y=d.height;
		if (legal(x,y)) {
			mx = x; my = y;
		}
		
		d = moveCat(cx, cy, mx, my);
		cx = d.width;
		cy = d.height;		
		
		ro.terminal = false;
		ro.r = 0;

		if (walls[mx][my] == 3)
		{
		  	ro.r = cheeseReward;
			if (random_cheese) changeCheese();
		}
		if ((cx==mx) && (cy==my))
		{ 
			ro.r += deathPenalty;
			ro.terminal = true;
		}
				
		o[0] = mx;
		o[1] = my;
		o[2] = cx;
		o[3] = cy;

		ro.o = o;

		if (episode_count>verbose_env_from) env_print(); else printf(".");
		
		return ro;
	}
	
	void env_cleanup()
	{
		printf("\n");

		for( int i = 0; i < row; i++)
			delete [] walls[i];
		
		delete [] walls;	
	
	}
	
	Dimension moveMouse(int action) 
	{
		Dimension d;
		int x=mx, y=my;
		
		if (action == 0) 
			y = my - 1;
		else if(action == 1){
			y = my - 1; x = mx + 1;}
		else if(action == 2)
			x = mx + 1;
		else if(action == 3){
			y = my + 1; x = mx + 1;}
		else if(action == 4)
			y = my + 1;
		else if(action == 5){
			y = my + 1; x = mx - 1;}
		else if(action == 6)
			x = mx - 1; 
		else if(action == 7){
			y = my - 1; x = mx - 1;}
		else{
			printf("Invalid action: %d\nExiting ... \n\n",action);
			exit(0);
		}
		d.width = x;
		d.height = y;
		return d;
	}

	Dimension moveCat(int catX, int catY, int mouseX, int mouseY) {
		Dimension d;

 		int pas_trouve = true;

		int x=catX, y=catY;
		if (drand48()>epsilon_cat)
		{
			// mouvement normal
			if (mouseX==catX)
				x = catX;
			else
				x += (mouseX - catX)/abs(mouseX-catX); // +/- 1 or 0
			if (mouseY==catY)
				y = catY;
			else
				y += (mouseY - catY)/abs(mouseY-catY); // +/- 1 or 0
			pas_trouve = (!legal(x, y)  || walls[x][y] == 2);

			// si ça ne marche pas, on essaie chaque direction séparément pour trouver qque chose qui rapproche sans être optimal à fond
			// on fait x ou y d'abord au hasard
			int x_first = (int)(drand48()*2);
			int y_first = (int)(1-x_first);

			if (pas_trouve)
			{
				x=catX; y=catY;
				if (mouseX==catX)
					x = catX;
				else
					x += x_first*((mouseX - catX)/abs(mouseX-catX));
				if (mouseY==catY)
					y = catY;
				else
					y += y_first*((mouseY - catY)/abs(mouseY-catY));
				pas_trouve = (!legal(x, y)  || walls[x][y] == 2);	
			}

			if (pas_trouve)
			{
				x=catX; y=catY;
				if (mouseX==catX)
					x = catX;
				else
					x += y_first*((mouseX - catX)/abs(mouseX-catX));
				if (mouseY==catY)
					y = catY;
				else
					y += x_first*((mouseY - catY)/abs(mouseY-catY));
				pas_trouve = (!legal(x, y)  || walls[x][y] == 2);	
			}
		}		
		// si tout ça échoue, on fait un mouvement au hasard	
		int tick=0;
		while (pas_trouve && tick < 100)
		{
			tick++;
			x=catX; y=catY;
			x += 1 - (int)(drand48()*3);
			y += 1 - (int)(drand48()*3);
			pas_trouve = (!legal(x, y)  || walls[x][y] == 2);
		}
		d.width = x;
		d.height = y;
		return d;
	}

// 	Dimension moveCat(int catX, int catY, int mouseX, int mouseY) {
// 		Dimension d;
// 		int x=catX, y=catY;
//  		
// 		int pas_trouve = true;
// 
// 		while (pas_trouve)
// 			{
// 			if (mouseX==catX)
// 				x = catX;
// 			else if (drand48()>=epsilon_cat)
// 				x += (mouseX - catX)/abs(mouseX-catX); // +/- 1 or 0
// 			else
// 				x += 1-(int) (drand48() *3);
// 
// 			if (mouseY==catY)
// 				y = catY;
// 			else if (drand48()>=epsilon_cat)
// 				y += (mouseY - catY)/abs(mouseY-catY); // +/- 1 or 0
// 			else
// 				y += 1-(int) (drand48() *3);
// 			
// 			pas_trouve = (!legal(x, y)  || walls[x][y] == 2);
// 			}
// 
// 		d.width = x;
// 		d.height = y;
// 		return d;
// 	}

	void setRandomPos()
	{
		int dist_mini_cat_mouse = 2;

		Dimension d = getRandomPos();
		while(walls[d.width][d.height] != 0)
			d = getRandomPos();

		cx = d.width;
		cy = d.height;			

		// il doit y avoir au moins dist_mini_cat_mouse entre le chat et la souris, au départ
		d = getRandomPos();
		while( (abs(d.width-cx)>dist_mini_cat_mouse && abs(d.height-cy)>dist_mini_cat_mouse)
				|| (walls[d.width][d.height] == 2)
				|| (walls[d.width][d.height] == 3))
			d = getRandomPos();
		mx = d.width;
		my = d.height;
	}

	Dimension getRandomPos() 
	{
		Dimension d;
		int x, y;
		x = (int)(drand48() * row);
		y = (int)(drand48() * col);
		while(!legal(x,y))
		{
			x = (int)(drand48() * row);
			y = (int)(drand48() * col);
		}
		d.width = x;
		d.height = y;
		return d;
	}

	void changeCheese()
	{
		removeCheese();
		setRandomCheese();
	}

	void removeCheese()
	{
		walls[chx][chy]=0 ;
	}

	void setRandomCheese()
	{
		int x, y;
		x = (int)(drand48() * row);
		y = (int)(drand48() * col);
		while( (walls[x][y] != 0) || ((x==cx) && (y==cy)) || ((x==mx) && (y==my)) )
		{
			x = (int)(drand48() * row);
			y = (int)(drand48() * col);
		}
		walls[x][y]=3;
		chx=x;
		chy=y;
	}	
	
	bool legal(int x, int y) 
	{
		return ((x>=0) && (x<row) && (y>=0) && (y<col)) && (walls[x][y] != 1);
	}

	void env_print()
	{
		usleep(tempo);
		printf("\n");
		printf("Episode = %d,  t = %d\n",episode_count,step_count);
		for(int j=0;j<(col+2);j++) printf("#");
		printf("\n");
		for(int i=0;i<row;i++){
			printf("#");
			for(int j=0;j<col;j++)
				if ((mx==i) && (my==j) && (cx==i) && (cy==j))
					printf("!");
				else if ((mx==i) && (my==j) && (walls[i][j]==3))
					printf("@");
				else if ((walls[i][j]==2) && (mx==i) && (my==j))
					printf("s");
				else if ((mx==i) && (my==j))
					printf("S");
				else if ((cx==i) && (cy==j))
					printf("C");
				else if (walls[i][j]==2)
					printf("*");
				else if (walls[i][j]==1)
					printf("#");
				else if (walls[i][j]==3)
					printf("O");
				else if ((mx==cx) && (my==cy))
					printf(".");
				else
					printf(" ");
			printf("#\n");
		}
		for(int j=0;j<(col+2);j++) printf("#");
		printf("\n");
	
		fflush(stdout);
	}
