/*
 * Cree a partir de Mines_common.cpp par Adam White (Fri Aug 05 2005.)
 * Auteur : Pierre Grard pierre.gerard@lipn.univ-paris13.fr
 *
 * Environnement de type Gridworld avec des mines, des obstacles et un but
 * Les grilles sont generees aleatoirement
 *
 * Observations : int*
 * - int[0] : colonne de l'agent dans la grille
 * - int[1] : ligne de l'agent dans la grille
 * Actions : int
 * - 0,1,2,3 pour quatre directions
 * Rcompense : float
 * - 10 si le but est atteint
 * - -1 si l'agent passe sur une mine
 * - 0 sinon
 *
 * Tache episodique : un episode se termine quand l'agent atteint le but
 * Chaque episode, le point de depart est choisi aleatoirement
 */

#include "Gridworld.h"

long step_count;
long episode_count;

int random_placement = false;  	// est-ce que le placement du fromage est aléatoire ?
				// à n'utiliser qu'avec les environnements sans fromage

long verbose_env_from = 490; // épisode à partir duquel l'environement devient verbose

int tempo = 50000;		// temporisation enre deux pas de temps, dans le cas verbose

int grid_size = 20;

void randStart()
{
	int r,c;
	if (random_placement)
	{
		r = (int)(drand48()*M.row);
		c = (int)(drand48()*M.col);
		
		while(env_map[r][c] != M.LAND)
		{
			r = (int)(drand48()*M.row);
			c = (int)(drand48()*M.col);
		}
	}
	else
	{
		r = 1;
		c = 1;
	}

	M.agentRow = M.startRow = r;
	M.agentColumn = M.startCol = c;
	env_map[M.startRow][M.startCol] = M.START;
}


void newRandStart()
{
	env_map[M.startRow][M.startCol] = M.LAND;
	randStart();

}

void gen_rand_map()
{
	srand48(0);

	int r,c;

	step_count=episode_count=0;

	for(int i=0;i< M.row; i++)
		for(int j=0;j< M.col; j++)
		{
			double rd = drand48();
			if (rd < 0.1)
				env_map[i][j] = M.OBSTACLE;
			else if (rd < 0.2)
				env_map[i][j] = M.MINE;
			else
				env_map[i][j] = M.LAND;
		}

	if (random_placement)
	{
		r = (int)(drand48()*M.row);
		c = (int)(drand48()*M.col);
	}
	else
	{
		r = M.row-2 ;
		c = M.col-2 ;
	}

	M.mx = r;
	M.my = c;
	env_map[r][c] = M.GOAL;

	randStart();	
}

bool legal(int x, int y)
{
//	if(x >= 0 && y >= 0 && x < M.row && y < M.col && env_map[x][y] != 3 && env_map[x][y] != 4)
	if(x >= 0 && y >= 0 && x < M.row && y < M.col && env_map[x][y] != M.OBSTACLE)
		return true;
	else
		return false;
}

Task_specification env_init()
{    
    M.START = 0;
    M.GOAL = 1;
    M.LAND = 2;
    M.OBSTACLE = 3;
    M.MINE = 4;
	
    M.row = M.col = grid_size;
    gen_rand_map();
    static char ts[1024];
    sprintf(ts,"1:e:2_[i,i]_[0,%d]_[0,%d]:1_[i]_[0,3]",M.row-1,M.col-1);

    env_print(); fflush(stdout);
    return ts;
}

Observation env_start()
{
	episode_count++;
	step_count++;

	env_terminal = false;

	newRandStart();

	o[0] = M.startRow ;
	o[1] = M.startCol;

	if (episode_count>verbose_env_from) env_print(); else printf("*"); fflush(stdout);

	return o;
}

Reward_observation env_step(Action a)
{
    step_count++;

    ro.o = getNextPosition(a); 
    ro.r = getReward();
	
    if(env_terminal) //end of episode?
        ro.terminal = true;
    else
        ro.terminal = false;

    if (episode_count>verbose_env_from) env_print(); else printf("."); fflush(stdout);

    return ro;
}

int* getPosition()
{
	o[0]=M.agentRow;
	o[1]=M.agentColumn;

	env_terminal = (env_map[M.agentRow][M.agentColumn] == M.GOAL);

        return o;
}

int* getNextPosition(int a)
{
    // When the move would result in hitting an obstacles, the agent simply doesn't move
	int oldR = M.agentRow; 
	int oldC = M.agentColumn ;

    if (a == 0)
        M.agentColumn = M.agentColumn - 1;
    else if (a == 1)
        M.agentColumn = M.agentColumn + 1;
    else if (a == 2)
        M.agentRow = M.agentRow - 1;
    else if (a == 3)
        M.agentRow = M.agentRow + 1;
    		
	if (!legal(M.agentRow,M.agentColumn))
	{
		M.agentRow = oldR;
		M.agentColumn = oldC;
	}
			
	return getPosition();
}

Reward getReward()
{
    if (env_map[M.agentRow][M.agentColumn] == M.GOAL)
        return 100;
    else if (env_map[M.agentRow][M.agentColumn] == M.MINE)
        return -10;
    else
        return -1;
}

void env_cleanup()
{
}

void env_print()
{
	usleep(tempo);
	printf("\n");
	printf("Episode = %d,  t = %d\n",episode_count,step_count); 

	for(int j=0;j<(M.col+2);j++) printf("#");
	printf("\n");
	for(int i=0;i<M.row;i++){
		printf("#");
		for(int j=0;j<M.col;j++)
			if ((M.agentRow==i) && (M.agentColumn==j))
				printf("O");
			else if (env_map[i][j]==M.START)
				printf("S");
			else if (env_map[i][j]==M.GOAL)
				printf("X");
			else if (env_map[i][j]==M.LAND)
				printf(" ");
			else if (env_map[i][j]==M.OBSTACLE)
				printf("#");
			else if (env_map[i][j]==M.MINE)
				printf("+");
		printf("#\n");
	}
	for(int j=0;j<(M.col+2);j++) printf("#");
	printf("\n");

	fflush(stdout);
}